@@ -51,7 +51,9 @@ sub Edit($)
 			$ENV{"P4EDITOR"} :
 			defined( $ENV{"EDITOR"} ) ?
 				$ENV{"EDITOR"} :
-				"vi";
+				$^O eq "MSWin32" ? 
+					"notepad" : 
+					"vi";
 
 	system( "$editor $filename" );
 }
@@ -95,7 +97,7 @@ sub OutputInfo($$)
 sub OutputError($)
 {
 	my ($self, $err) = @_;
-	print( $err );
+	print( STDERR $err );
 }
 
 
@@ -103,8 +105,19 @@ sub OutputError($)
 sub OutputStat
 {
 	my ($self, $hash ) = @_;
+
 	foreach my $key ( sort keys %$hash )
 	{
+	    if ( ref( $hash->{ $key } ) )
+	    {
+		# Nested element
+		print( "... $key\n" );
+		foreach my $item ( @{ $hash->{ $key } } )
+		{
+		    print( "... ... $item\n" );
+		}
+	    }
+
 	    print("... $key ", $hash->{ $key }, "\n" );
 	}
 	print("\n");
@@ -114,7 +127,7 @@ sub OutputStat
 sub OutputText($$)
 {
 	my ($self, $text, $len ) = @_;
-	print( $text );		# Ignore length unless it becomes an issue
+	printf( "%*s", $len, $text );
 }
 
 
@@ -128,6 +141,30 @@ sub Prompt($)
 	return <>;
 }
 
+
+#
+# Function to diff two files. This default implementation does nothing of
+# great use. Users should consider using the Algorithm::Diff module to get
+# real diff listings.
+#
+sub Diff($$$$)
+{
+    my $self 		= shift;
+    my $f1		= shift;
+    my $f2		= shift;
+    my $diffFlags	= shift;
+    my $differ		= shift;
+
+    if ( $differ )
+    {
+	print( "$f1 and $f2 differ\n" );
+    }
+    else
+    {
+	print( "$f1 and $f2 are the same\n" );
+    }
+}
+
 1;
 
 __END__
@@ -168,11 +205,69 @@ derive a subclass from P4::UI and override the appropriate methods.
 
 =over 4
 
-=item C<UI::new()>
+=item C<new()>
 
 	Constructor. The only method of this class you call
 	directly.
 
+=item C<Edit( $filename )>
+
+	Method called when the Perforce server wants the client to
+	edit a specform ( changespec, jobspec, clientspec etc.). 
+	The default implementation simply invokes your chosen editor
+	by checking the environment variables P4EDITOR and EDITOR
+	before falling back on vi/notepad and hoping they're there.
+
+=item C<ErrorPause( $message )>
+
+	Called to alert the user to an error message which they
+	must acknowledge before continuing. Default implementation
+	prints the message on stdout and prompts the user to "Hit
+	Return to continue".
+
+=item C<InputData()>
+
+	Used to read information directly from the user. Called
+	in response to the "-i" flag to many Perforce commands.
+	i.e. "p4 submit -i".
+
+=item C<OutputInfo( $level, $data )>
+
+	Writes data to stdout prefixed by $level occurances of 
+	"..." 
+
+=item C<OutputError( $error )>
+
+	Blurt an error to stderr. The error message arrives 
+	as preformatted text. 
+
+=item C<OutputStat( $hashref )>
+
+	Print the output of a command in tagged format. The tagged
+	data is passed as a hash reference for ease of use. Multi 
+	line fields (such as mappings) are presented as array
+	references within the hash. i.e. in a clientspec, the 
+	View field in the form is normally transmitted by Perforce 
+	as a sequence of fields: "View0", "View1", etc. P4::UI gets
+	these as a single "View" member of the hash which is itself 
+	an array reference containing the view records in order.
+
+=item C<OutputText( $text, $length )>
+
+	Prints $length bytes of $text on STDOUT
+
+=item C<Prompt( $prompt )>
+
+	Prints the prompt string $prompt and then reads a line
+	of input from the user returning the input line.
+
+=item C<Diff( $f1, $f2, $flags, $differ )>
+
+	Diff two files. The default implementation of this method
+	does nothing of any great use. It's intended to be overridden
+	in subclasses.
+	
+
 =back
 
 =head1 LICENCE
